"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var util = require("../lib/util");
function pollUntil() {
    var allArgs = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        allArgs[_i] = arguments[_i];
    }
    var poller = allArgs[0], args = allArgs[1], timeout = allArgs[2], pollInterval = allArgs[3];
    if (typeof args === 'number') {
        pollInterval = timeout;
        timeout = args;
        args = [];
    }
    args = args || [];
    pollInterval = pollInterval || 67;
    return function () {
        var session = this.session;
        var originalTimeout;
        return session.getExecuteAsyncTimeout().then(function () {
            var resultOrError;
            function storeResult(result) {
                resultOrError = result;
            }
            function finish() {
                if (resultOrError instanceof Error) {
                    throw resultOrError;
                }
                if (resultOrError == null) {
                    var error = new Error('Polling timed out with no result');
                    error.name = 'ScriptTimeout';
                    throw error;
                }
                return resultOrError;
            }
            function cleanup() {
                if (!isNaN(originalTimeout)) {
                    return session.setExecuteAsyncTimeout(originalTimeout).then(finish);
                }
                return finish();
            }
            if (!isNaN(timeout)) {
                originalTimeout = arguments[0];
            }
            else {
                timeout = arguments[0];
            }
            return session.setExecuteAsyncTimeout(timeout)
                .then(function () {
                /* jshint maxlen:140 */
                return session.executeAsync(/* istanbul ignore next */ function (poller, args, timeout, pollInterval, done) {
                    /* jshint evil:true */
                    poller = new Function(poller);
                    var endTime = Number(new Date()) + timeout;
                    (function poll() {
                        var result = poller.apply(this, args);
                        /*jshint evil:true */
                        if (result != null) {
                            done(result);
                        }
                        else if (Number(new Date()) < endTime) {
                            setTimeout(poll, pollInterval);
                        }
                        else {
                            done(null);
                        }
                    })();
                }, [util.toExecuteString(poller), args, timeout, pollInterval]);
            })
                .then(storeResult, storeResult)
                .then(cleanup, cleanup);
        });
    };
}
exports.default = pollUntil;
//# sourceMappingURL=pollUntil.js.map