"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Task_1 = require("@dojo/core/async/Task");
var lang_1 = require("@dojo/core/lang");
/**
 * Creates a promise that resolves itself after `ms` milliseconds.
 *
 * @param ms Time until resolution in milliseconds.
 */
function sleep(ms) {
    var timer;
    return new Task_1.default(function (resolve) {
        timer = setTimeout(function () {
            resolve();
        }, ms);
    }, function () { return clearTimeout(timer); });
}
exports.sleep = sleep;
/**
 * Annotates the method with additional properties that provide guidance to [[Command]] about
 * how the method interacts with stored context elements.
 */
function forCommand(fn, properties) {
    return lang_1.mixin(fn, properties);
}
exports.forCommand = forCommand;
/**
 * Converts a function to a string representation suitable for use with the `execute` API endpoint.
 */
function toExecuteString(fn) {
    if (typeof fn === 'function') {
        // If someone runs code through Istanbul in the test runner, inline functions that are supposed to execute
        // on the client will contain code coverage variables that will cause script execution failure. These
        // statements are very simple and are generated in a consistent manner, so we can get rid of them easily
        // with a regular expression
        fn = fn.toString().replace(/\b__cov_[^,;]+[,;]/g, '');
        fn = 'return (' + fn + ').apply(this, arguments);';
    }
    return fn;
}
exports.toExecuteString = toExecuteString;
/**
 * Removes the first line of a stack trace, which in V8 is the string representation of the object holding the stack
 * trace (which is garbage for captured stack traces).
 */
function trimStack(stack) {
    return stack.replace(/^[^\n]+/, '');
}
exports.trimStack = trimStack;
function applyMixins(derivedCtor, baseCtors, includePrivates) {
    if (includePrivates === void 0) { includePrivates = true; }
    baseCtors.forEach(function (baseCtor) {
        Object.getOwnPropertyNames(baseCtor.prototype).forEach(function (name) {
            if (name !== 'constructor' && (includePrivates || name.charAt(0) !== '_')) {
                derivedCtor.prototype[name] = baseCtor.prototype[name];
            }
        });
    });
}
exports.applyMixins = applyMixins;
//# sourceMappingURL=util.js.map